package at.ac.tuwien.dbai.alternation.runtime;

import java.util.Map;

/**
 * This class uses tabled evaluation to simulate an alternating program.
 *
 * @author Wolfgang Dvorak
 *
 * @param <GInputtape> Container class for the input variables
 * @param <GWorktape> Container class for the work variables
 */
public class ATMacyclic<GInputtape,GWorktape extends InterfaceWorktape<GInputtape>> implements ATM<GInputtape, GWorktape> {

	private Map<Byte,State<GInputtape,GWorktape>> states;
	private byte startState;
	private GWorktape worktape;

	/**
	 * Data structure for the tabled evaluation
	 * the tree stucture saves a little bit more information which is usefull in some cases
	 */
	private ComputationTree<GWorktape> computationTree;

	/**
	 * constructor
	 * @param states Map with states
	 * @param worktape a (empty) worktape
	 * @param startState stateid from the state the computation should start with
	 */
	public ATMacyclic(Map<Byte,State<GInputtape,GWorktape>> states,GWorktape worktape, byte startState){
		this.states=states;
		this.startState=startState;
		this.worktape=worktape;
	}
	/* (non-Javadoc)
	 * @see at.ac.tuwien.dbai.staff.dvorak.atm.ATM#compute(GInputtape)
	 */
	public boolean compute(GInputtape input){
		byte stateId=startState;

		computationTree=new ComputationTree<GWorktape>();
		for(State<GInputtape,GWorktape> state: states.values()){
			state.setInput(input);
		}
		worktape.reset(input);

		ComputationNode<GWorktape> rootNode = new ComputationNode<GWorktape>(stateId,worktape);
		computationTree.setRoot(rootNode);

		return computeNode(rootNode);
	};

	/**
	 * computes the result of a ComputationNode
	 * @param node ComputationNode
	 * @return
	 */
	private boolean computeNode(ComputationNode<GWorktape> node) {
		ComputationNode<GWorktape> node2;

		// search in computationTree (tabled evaluation)
		if (computationTree.containsKey(node)){
			return computationTree.get(node);
		}

		ResultTuple<GWorktape> result;
		GWorktape worktape=(GWorktape)node.worktape.clone();

		result=states.get(node.state).compute(worktape);

		if (result.accept){
			computationTree.put(node,true);//tabled evaluation
			return true;
		}

		if (result.reject){
			computationTree.put(node,false);//tabled evaluation
			return false;
		}

		if (result.forall){
			/*
			 * What to do if there are no successors ? - default behaviour is accepting
			 * if (result.successors.isEmpty() ){return false;}
			 */
			node.setQuantifier(true);
			for (ResultTuple.AtmConfiguration<GWorktape> work : result.successors) {
				node2=new ComputationNode<GWorktape>(work.state,work.worktape);
				computationTree.addEdge(node,node2);
				if (!computeNode(node2)){
					computationTree.put(node,false);//tabled evaluation
					return false;
				}
			}
			computationTree.put(node,true);//tabled evaluation
			return true;
		} else {
			node.setQuantifier(false);
			for (ResultTuple.AtmConfiguration<GWorktape> work : result.successors) {
				node2=new ComputationNode<GWorktape>(work.state,work.worktape);
				computationTree.addEdge(node,node2);
				if (computeNode(node2)){
					computationTree.put(node,true);//tabled evaluation
					return true;
				}
			}
			computationTree.put(node,false);//tabled evaluation
			return false;
		}
	}
	/* (non-Javadoc)
	 * @see at.ac.tuwien.dbai.staff.dvorak.atm.ATM#getComputationTree()
	 */
	public ComputationTree<GWorktape> getComputationTree(){
		return this.computationTree;
	}
}